/***************************************************************************
 *   Copyright (C) 2004 by Jens Luetkens                                   *
 *   j.luetkens@limitland.de                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qpainter.h>
#include <qtooltip.h>
#include <qbutton.h>
#include <klocale.h>
#include <kpixmap.h>
#include <kpixmapeffect.h>

/*#include "Ridge.h"*/
#include "RidgeClient.h"
#include "RidgeButton.h"

using namespace RIDGE;

static unsigned char help_bits[] =			{ 0x18, 0x18, 0x00, 0x1e, 0x10, 0x10, 0x10, 0x10 };
static unsigned char sticky_bits[] =		{ 0x3c, 0x7e, 0xe7, 0xc3, 0xc3, 0xe7, 0x7e, 0x3c };
static unsigned char unsticky_bits[] =		{ 0x3c, 0x7e, 0xff, 0xff, 0xff, 0xff, 0x7e, 0x3c };
static unsigned char iconify_bits[] =		{ 0x00, 0x00, 0x3c, 0x3c, 0x3c, 0x3c, 0x00, 0x00 };
static unsigned char maximize_bits[] =		{ 0xff, 0xff, 0xc3, 0xc3, 0xc3, 0xc3, 0xff, 0xff };
static unsigned char unmaximize_bits[] =	{ 0x00, 0x00, 0xff, 0xff, 0xc3, 0xc3, 0xff, 0xff };
static unsigned char close_bits[] =			{ 0xc3, 0xe7, 0x7e, 0x3c, 0x3c, 0x7e, 0xe7, 0xc3 };
static unsigned char empty_bits[] =			{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
static const int buttonSize = 				8;
static const int halfbuttonSize = 			4;

RidgeButton::RidgeButton(RidgeClient * parent, const char *name, const QString& tip, ButtonType type)
		: QButton(parent->widget(), name),
		client_		(parent),
		mouseOver_  (false),
		mouseDown_  (false)
{
	
	switch (type) {

		case ButtonHelp:
			bitmap_ = QBitmap(buttonSize, buttonSize, help_bits, true);
			break;

		case ButtonOnAllDesktops:
			if ( parent->isOnAllDesktops() ) {
				bitmap_ = QBitmap(buttonSize, buttonSize, unsticky_bits, true);
			}
			else {
				bitmap_ = QBitmap(buttonSize, buttonSize, sticky_bits, true);
			}
			break;

		case ButtonIconify:
			bitmap_ = QBitmap(buttonSize, buttonSize, iconify_bits, true);
			break;

		case ButtonMaximize:
			if ( parent->widget()->isMaximized() ) {
				bitmap_ = QBitmap(buttonSize, buttonSize, unmaximize_bits, true);
			}
			else {
				bitmap_ = QBitmap(buttonSize, buttonSize, maximize_bits, true);
			}
			break;

		case ButtonClose:
			bitmap_ = QBitmap(buttonSize, buttonSize, close_bits, true);
			break;
		
		default:
			bitmap_ = QBitmap(buttonSize, buttonSize, empty_bits, true);
			break;
	}
	
	bitmap_.setMask(bitmap_);

	setSizePolicy(QSizePolicy(QSizePolicy::Minimum, QSizePolicy::Minimum));
	setBackgroundMode(NoBackground);

    QToolTip::add(this, tip);

	repaint();
}

RidgeButton::~RidgeButton()
{
	// Empty.
}

//
// setOnAllDesktops(bool b)
//
void RidgeButton::setOnAllDesktops(bool b)
{
	if ( b ) {
		bitmap_ = QBitmap(buttonSize, buttonSize, unsticky_bits, true);
	}
	else {
		bitmap_ = QBitmap(buttonSize, buttonSize, sticky_bits, true);
	}
	bitmap_.setMask(bitmap_);

	repaint();
}

//
// slotMaximizeChange(bool b)
//
void RidgeButton::setMaximized(bool b)
{
	if ( b ) {
		bitmap_ = QBitmap(buttonSize, buttonSize, unmaximize_bits, true);
	}
	else {
		bitmap_ = QBitmap(buttonSize, buttonSize, maximize_bits, true);
	}
	bitmap_.setMask(bitmap_);

	repaint();
}

//
// enterEvent(QEvent * e)
//
void RidgeButton::enterEvent(QEvent * e)
{
	mouseOver_ = true;
	repaint();
	QButton::enterEvent(e);
}

//
// leaveEvent(QEvent * e)
//
void RidgeButton::leaveEvent(QEvent * e)
{
	mouseOver_ = false;
	repaint();
	QButton::leaveEvent(e);
}

//
// mousePressEvent(QMouseEvent* e)
//
void RidgeButton::mousePressEvent(QMouseEvent* e)
{
    lastMouse_ = e->button();
    // pass on event after changing button to LeftButton
    QMouseEvent me(e->type(), e->pos(), e->globalPos(),
                   LeftButton, e->state());
    QButton::mousePressEvent(&me);
}

//
// mouseReleaseEvent(QMouseEvent* e)
//
void RidgeButton::mouseReleaseEvent(QMouseEvent* e)
{
    lastMouse_ = e->button();
    // pass on event after changing button to LeftButton
    QMouseEvent me(e->type(), e->pos(), e->globalPos(),
                   LeftButton, e->state());
    QButton::mouseReleaseEvent(&me);
}

//
// paintEvent(QPaintEvent *)
//
void RidgeButton::paintEvent(QPaintEvent *)
{
	//Ridge *client = (Ridge *) parentWidget();

	QPainter p(this);
	QRect buttonRect(rect());
    bool active = client_->isActive();

	//tint colors for titlebar
    titleBarColor = client_->options()->color(KDecoration::ColorTitleBar, active);
	
	if ( client_->m_hasOutline ) {
		
		client_->frameRidgeRect(buttonRect, &p);
		client_->insetRect(&buttonRect, 1);

	}
	client_->decorateRidgeRect(buttonRect, &p, titleBarColor, true);

	// Draw bitmap icon.
	QPoint center(buttonRect.center());

	p.setBrush(NoBrush);
	p.setPen(client_->options()->color(KDecoration::ColorFont, mouseOver_));
	p.drawPixmap(center.x() - halfbuttonSize + 1, center.y() - halfbuttonSize + 1, bitmap_);

}

