/***************************************************************************
 *   Copyright (C) 2004 by Jens Luetkens                                   *
 *   j.luetkens@limitland.de                                               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>

#include <qlayout.h>
#include <qlabel.h>
#include <qpainter.h>
#include <qimage.h>
#include <qtooltip.h>

#include <kconfig.h>
#include <kstandarddirs.h>
#include <kapplication.h>
#include <klocale.h>
#include <kpixmap.h>
#include <kpixmapeffect.h>

#include <kdecoration.h>
#include <kdecorationfactory.h>

#include "RidgeClient.h"
#include "RidgeButton.h"


using namespace RIDGE;

// Default button layout
const char default_left[]  = "M";
const char default_right[] = "HIAX";

//
// Constructor
//
RidgeClient::RidgeClient( KDecorationBridge *b, KDecorationFactory *f )
	    : KDecoration( b, f ),
		mainLayout_(0),
		topSpacer_(0),
		menuIcon(0),
		activeTile(0),
		inactiveTile(0),
		titleWidth(0),
		topHeight(0),
		fontHeight(0),
		textMargin(0),
		boxHeight(0),
		titleBarColor(0,0,0),
		titleBlendColor(0,0,0),
		hasMenu(false)
{
}

//
// Destructor
//
RidgeClient::~RidgeClient()
{

    deletePixmaps();

    for (int n=0; n<ButtonCount; n++) {
        if (m_buttons[n]) delete m_buttons[n];
    }

}

//
// init(void)
//
void RidgeClient::init( void )
{
    createMainWidget(WResizeNoErase | WRepaintNoErase);
    widget()->installEventFilter(this);

    widget()->setBackgroundMode(NoBackground);
	
	readConfig();

    createLayout();

    createPixmaps();

}

//
// readConfig()
//
bool RidgeClient::readConfig()
{

	QColor defaultAColor(74,84,95);
	QColor defaultIColor(58,67,75);

    // create a config object
    KConfig config("kwinridgerc");
    config.setGroup("General");

    m_hasBorder				= config.readBoolEntry("hasBorder", true);
    m_borderSize			= config.readNumEntry("borderSize", 3);

    m_hasOffset				= config.readBoolEntry("hasOffset", true);
    m_offsetSize			= config.readNumEntry("offsetSize", 5);

    m_hasDecoration 		= config.readBoolEntry("hasDecoration", true);
	m_decorationTheme		= config.readEntry("decorationTheme", "Gradient");

    m_hasOutline			= config.readBoolEntry("hasOutline", true);
	m_outlineAColor			= config.readColorEntry("outlineAColor", &defaultAColor);
	m_outlineIColor			= config.readColorEntry("outlineIColor", &defaultIColor);

	someGeometry();

    return true;
}

//
// createPixmaps()
//
void RidgeClient::createPixmaps()
{

	//
	// if decorating 
	//
	QPainter	painter;
   	KPixmap		tempPixmap;

	QColor activeTitleBarColor 		= options()->colorGroup(KDecoration::ColorTitleBar, true).background();
	QColor inactiveTitleBarColor 	= options()->colorGroup(KDecoration::ColorTitleBar, false).background();
	QColor activeTitleBlendColor 	= options()->colorGroup(KDecoration::ColorTitleBlend, true).background();
	QColor inactiveTitleBlendColor 	= options()->colorGroup(KDecoration::ColorTitleBlend, false).background();

	if ( m_hasOutline )
		tempPixmap.resize(1, boxHeight-1);
	else
		tempPixmap.resize(1, boxHeight);

	KPixmapEffect::gradient(tempPixmap,
                	  activeTitleBarColor,
                	  activeTitleBlendColor,
                	  KPixmapEffect::VerticalGradient);

	activeTile = new QPixmap(1, tempPixmap.height() );
	painter.begin(activeTile);
	painter.drawPixmap(0, 0, tempPixmap);
	painter.end();

	KPixmapEffect::gradient(tempPixmap,
                	  inactiveTitleBarColor,
                	  inactiveTitleBlendColor,
                	  KPixmapEffect::VerticalGradient);

	inactiveTile = new QPixmap(1, tempPixmap.height() );
	painter.begin(inactiveTile);
	painter.drawPixmap(0, 0, tempPixmap);
	painter.end();
	
	//
	// if hasMenu
	//
	menuIcon = icon().pixmap(QIconSet::Small, QIconSet::Normal);
	menuIcon.convertFromImage( menuIcon.convertToImage().smoothScale(fontHeight-1, fontHeight-1));

}

//
// deletePixmaps()
//
void RidgeClient::deletePixmaps()
{

	delete activeTile;
	delete inactiveTile;
	
}

//
// eventFilter( QObject *obj, QEvent *e )
//
bool RidgeClient::eventFilter( QObject *obj, QEvent *e )
{
    if (obj != widget()) return false;

    switch (e->type()) {
      case QEvent::MouseButtonDblClick: {
          mouseDoubleClickEvent(static_cast<QMouseEvent *>(e));
          return true;
      }
      case QEvent::MouseButtonPress: {
          processMousePressEvent(static_cast<QMouseEvent *>(e));
          return true;
      }
      case QEvent::Paint: {
          paintEvent(static_cast<QPaintEvent *>(e));
          return true;
      }
      case QEvent::Resize: {
          resizeEvent(static_cast<QResizeEvent *>(e));
          return true;
      }
      case QEvent::Show: {
          showEvent(static_cast<QShowEvent *>(e));
          return true;
      }
      default: {
          return false;
      }
    }

    return false;
}

//
// paintEvent(QPaintEvent *)
//
void RidgeClient::paintEvent(QPaintEvent *)
{
	QPainter p(widget());
	QFontMetrics fm(options()->font(isActive(), false));
	
	titleBarColor = options()->color(KDecoration::ColorTitleBar, isActive());
	titleBlendColor = options()->color(KDecoration::ColorTitleBlend, isActive());


		QRect titleRect(topSpacer_->geometry());
		titleRect.setHeight(boxHeight);
		titleWidth = QMIN(fm.width(caption()), titleRect.width());
		titleWidth += textMargin + textMargin + textMargin;
		if ( hasMenu ) {
			titleWidth += textMargin + menuIcon.width();
		}
		titleRect.setWidth(titleWidth);


		if ( isShade() ) {

			setMask(titleRect);

		}	
		else {

			QRect windowRect(0,0,width(),height());

			// window background
			if ( m_hasOffset )
				windowRect.setTop(m_offsetSize);
			QRegion mask(windowRect);
			if ( m_hasOutline ) {
				frameRidgeRect(windowRect, &p);
				insetRect(&windowRect, 1);
			}

			if ( m_decorationTheme == "Gradient" )
				p.fillRect(windowRect, titleBlendColor);
			else 
				decorateRidgeRect(windowRect, &p, titleBlendColor, true);


			// top Layout Background
			if ( m_hasOutline )
				windowRect.setHeight(boxHeight-1);
			else
				windowRect.setHeight(boxHeight);
			decorateRidgeRect(windowRect, &p, titleBlendColor, false);

			// Mask without buttons
			for (int n=0; n<ButtonCount; n++) {
        		if (m_buttons[n]) {
					mask += m_buttons[n]->frameGeometry();
				}
			}

			mask += titleRect;
			setMask(mask);

		}

		// title Rect
		if ( m_hasOutline ) {
			frameRidgeRect(titleRect, &p);
			insetRect(&titleRect, 1);
		}
		decorateRidgeRect(titleRect, &p, titleBarColor, true);
		insetRect(&titleRect, textMargin);


		// draw icon
		if ( hasMenu ) {
	    	p.drawPixmap(titleRect.left(), titleRect.top(), menuIcon);
			titleRect.moveBy(menuIcon.width()+textMargin,0);
			titleRect.setWidth(titleRect.width()-menuIcon.width()-textMargin);
		}
		titleRect.setTop(titleRect.top()-(textMargin/2));
		titleRect.setHeight(titleRect.height()+(textMargin/2));

		//draw text
		// isTool ?
		p.setFont(options()->font(isActive(), false));
		p.setPen(options()->color(KDecoration::ColorFont, isActive()));
		p.drawText(titleRect, AlignLeft, caption());

	
}

//
// decorateRidgeRect( QRect rect, QPainter * p, QColor backgroundColor, bool fill )
//
void RidgeClient::decorateRidgeRect( QRect rect, QPainter * p, QColor backgroundColor, bool fill )
{
	if ( m_hasDecoration ) {
		if ( m_decorationTheme == "Gradient" ) {
				p->drawTiledPixmap(rect, isActive() ? *activeTile : *inactiveTile);
			
		}
		else {
			QColor lightColor = backgroundColor.light(120);
			QColor darkColor = backgroundColor.dark(120);

			if ( fill )
				p->fillRect(rect, backgroundColor);
			
			//draw long bevelling lines
			p->setPen(darkColor);
			p->drawLine(rect.left()+1, rect.bottom(), rect.right(), rect.bottom());
			p->drawLine(rect.right(), rect.bottom(),rect.right(), rect.top()+1);
			p->drawLine(rect.left()+1, rect.bottom()-2, rect.left()+1, rect.top()+1);
			p->drawLine(rect.left()+1, rect.top()+1, rect.right()-2, rect.top()+1);
			p->setPen(lightColor);
			p->drawLine(rect.left(), rect.bottom()-1, rect.left(), rect.top());
			p->drawLine(rect.left(), rect.top(), rect.right()-1, rect.top());
			p->drawLine(rect.left()+2, rect.bottom()-1, rect.right()-1, rect.bottom()-1);
			p->drawLine(rect.right()-1, rect.bottom()-1,rect.right()-1, rect.top()+2);
		}
	}
	else {
		p->fillRect(rect, backgroundColor);
	}
}

//
// frameRidgeRect( QRect rect, QPainter * p 
//
void RidgeClient::frameRidgeRect( QRect rect, QPainter * p )
{

	if ( isActive() )
		p->setPen(m_outlineAColor);
	else
		p->setPen(m_outlineIColor);

	p->drawRect(rect);

}

//
// insetRect( QRect * rect, int inset )
//
void RidgeClient::insetRect( QRect * rect, int inset )
{

	rect->setWidth(rect->width()-inset-inset);
	rect->setHeight(rect->height()-inset-inset);
	rect->moveBy(inset,inset);

}

//
// mouseDoubleClickEvent(QMouseEvent * e)
//
void RidgeClient::mouseDoubleClickEvent(QMouseEvent * e)
{
	if (topSpacer_->geometry().contains(e->pos()))
        titlebarDblClickOperation();
}

void RidgeClient::captionChange(void)
{
    widget()->repaint(topSpacer_->geometry(), false);
}

//
// desktopChange()
//
void RidgeClient::desktopChange()
{
    bool s = isOnAllDesktops();
    if (m_buttons[ButtonOnAllDesktops])
    {
        m_buttons[ButtonOnAllDesktops]->update();
        m_buttons[ButtonOnAllDesktops]->setOnAllDesktops(s);
        QToolTip::add(m_buttons[ButtonOnAllDesktops], s ? i18n("Not On All Desktops") : i18n("On All Desktops"));
    }
}


//
// maximizeChange()
//
void RidgeClient::maximizeChange()
{
    widget()->repaint( false );
}

//
// activeChange()
//
void RidgeClient::activeChange()
{
	widget()->repaint(false);
	repaintButtons();
}

//
// resize (const QSize &size)
//
void RidgeClient::resize (const QSize &size)
{
    widget()->resize(size);
}

//
// borders(int &l, int &r, int &t, int &b) const
//
void RidgeClient::borders(int &l, int &r, int &t, int &b) const
{

	t = topHeight;
	if ( m_hasBorder )
		b = l = r = m_borderSize;
	else
		b = l = r = 0;
		
}

//
// createButtons()
//
void RidgeClient::createButtons(QBoxLayout *layout, const QString& buttons)
{

	for (unsigned int i = 0; i < buttons.length(); ++i)
	{
		//RidgeButton * b = 0;
		switch (buttons[i].latin1())
		{
		case 'H': // Help
			if ((!m_buttons[ButtonHelp]) && providesContextHelp()){
				m_buttons[ButtonHelp] = new RidgeButton(this, "help", i18n("Help"), ButtonHelp);
				connect(m_buttons[ButtonHelp], SIGNAL(clicked()), this, SLOT(contextHelp()));
				layout->addWidget(m_buttons[ButtonHelp], 0, Qt::AlignTop);
				layout->setStretchFactor(m_buttons[ButtonHelp], 0);
				m_buttons[ButtonHelp]->setFixedSize(boxHeight, boxHeight);
            }
			break;
		case 'M': // Menu
			hasMenu = true;
			break;
		case 'S': // On All Desktops (Sticky)
			m_buttons[ButtonOnAllDesktops] = new RidgeButton(this, "onalldesktops", i18n("On All Desktops"), ButtonOnAllDesktops);
			connect(m_buttons[ButtonOnAllDesktops], SIGNAL(clicked()), this, SLOT(toggleOnAllDesktops()));
			layout->addWidget(m_buttons[ButtonOnAllDesktops], 0, Qt::AlignTop);
			layout->setStretchFactor(m_buttons[ButtonOnAllDesktops], 0);
			m_buttons[ButtonOnAllDesktops]->setFixedSize(boxHeight, boxHeight);
			break;
		case 'I': // Minimize
			if (isMinimizable()){
				m_buttons[ButtonIconify] = new RidgeButton(this, "iconify", i18n("Minimize"), ButtonIconify);
				connect(m_buttons[ButtonIconify], SIGNAL(clicked()), this, SLOT(minimize()));
				layout->addWidget(m_buttons[ButtonIconify], 0, Qt::AlignTop);
				layout->setStretchFactor(m_buttons[ButtonIconify], 0);
				m_buttons[ButtonIconify]->setFixedSize(boxHeight, boxHeight);
			}
			break;
		case 'A': // Maximize
			if (isMaximizable()){
				m_buttons[ButtonMaximize] = new RidgeButton(this, "maximize", i18n("Maximize"), ButtonMaximize);
				connect(m_buttons[ButtonMaximize], SIGNAL(clicked()), this, SLOT(maxButtonPressed()));
				layout->addWidget(m_buttons[ButtonMaximize], 0, Qt::AlignTop);
				layout->setStretchFactor(m_buttons[ButtonMaximize], 0);
				m_buttons[ButtonMaximize]->setFixedSize(boxHeight, boxHeight);
			}
			break;
		case 'X': // Close
			if (isCloseable()){
				m_buttons[ButtonClose] = new RidgeButton(this, "close", i18n("Close"), ButtonClose);
				connect(m_buttons[ButtonClose], SIGNAL(clicked()), this, SLOT(closeWindow()));
				layout->addWidget(m_buttons[ButtonClose], 0, Qt::AlignTop);
				layout->setStretchFactor(m_buttons[ButtonClose], 0);
				m_buttons[ButtonClose]->setFixedSize(boxHeight, boxHeight);
			}
			break;
		case '_': // Spacer item
			layout->addSpacing(m_offsetSize);
			break;
		}
	}

}

//
// someGeometry()
//
void RidgeClient::someGeometry() 
{
	
	QFontMetrics fm(options()->font(isActive(), false));
	
	textMargin = QMAX(3,fm.height() / 7);
	fontHeight = QMAX(8, fm.height());
	boxHeight = fontHeight + textMargin + textMargin;

	// must be a multiple of 2
	if (! boxHeight%2==0 )
		boxHeight += 1;

	if ( m_hasOffset ) {
		topHeight = boxHeight + m_offsetSize;
	}
	else {
		topHeight = boxHeight;
	}

}

//
// createLayout()
//
void RidgeClient::createLayout()
{

	// mainLayout:
	//  ____________________________________
	// |  |                              |  |
	// |Xo|     topSpacer                |v^| <--- topLayout
	// |__|______________________________|__|
	// | |                                | |
	// | |                                | |
	// | |     windowWrapper              | |
	// | |                                | | <--- midLayout
	// | |                                | |
	// | |                                | |
	// | |________________________________| |
	// |____________________________________|
	

	delete mainLayout_;
	mainLayout_		= new QVBoxLayout(widget(), 0, 0);

	topSpacer_		= new QSpacerItem(0,topHeight,QSizePolicy::Expanding,QSizePolicy::Fixed);

	QHBoxLayout * topLayout;
	if ( m_hasOffset )
		topLayout   = new QHBoxLayout(mainLayout_, m_offsetSize, 0);
	else 
		topLayout   = new QHBoxLayout(mainLayout_, 0, 0);
	
	// left offset
	if ( m_hasOffset )
		topLayout->addSpacing(m_offsetSize);

    // sizeof(...) is calculated at compile time
    memset(m_buttons, 0, sizeof(RidgeButton *) * ButtonCount);

	// Add left-side buttons.
	createButtons(topLayout, options()->customButtonPositions() ? options()->titleButtonsLeft() : QString(default_left));

	// the title spacer
	topLayout->addItem(topSpacer_);
	// space to the right of title
	topLayout->addSpacing(textMargin+textMargin+textMargin+textMargin);
	if ( hasMenu ) 
		topLayout->addSpacing(fontHeight);
	if ( m_hasOffset )
		topLayout->addSpacing(m_offsetSize);
	
	// Add right-side buttons.
	createButtons(topLayout, options()->customButtonPositions() ? options()->titleButtonsRight() : QString(default_right));

	// right offset
	if ( m_hasOffset )
		topLayout->addSpacing(m_offsetSize);

	// -------------------------------------------------------------------

	QHBoxLayout * midLayout   = new QHBoxLayout(mainLayout_, 0, 0);

	// left border
	if ( m_hasBorder )
		midLayout->addSpacing(m_borderSize);
	
	
	// window wrapper
    if( isPreview())
        midLayout->addWidget(new QLabel( i18n( "<center>Ridge preview</center>" ), widget()) );
    else
        midLayout->addItem( new QSpacerItem( 0, 0 ));
	
	// right border
	if ( m_hasBorder )
		midLayout->addSpacing(m_borderSize);

	// -------------------------------------------------------------------

	// bottom border
	if ( m_hasBorder )
		mainLayout_->addSpacing(m_borderSize);

	// -------------------------------------------------------------------

	// Make sure that topLayout doesn't stretch - midLayout should take
	// all spare space.
	mainLayout_->setStretchFactor(topLayout, 0);
	mainLayout_->setStretchFactor(midLayout, 1);

}

//
// repaintButtons(void)
//
void RidgeClient::repaintButtons(void)
{
	//force all buttons to repaint
    for (int n=0; n<ButtonCount; n++)
        if (m_buttons[n]) m_buttons[n]->repaint(false);

}

//
// maxButtonPressed()
//
void RidgeClient::maxButtonPressed()
{
    if (m_buttons[ButtonMaximize]) {
        switch (m_buttons[ButtonMaximize]->lastMousePress())
        {
          case MidButton:
              maximize(maximizeMode() ^ MaximizeVertical );
              maximizeChange();
              break;
          case RightButton:
              maximize(maximizeMode() ^ MaximizeHorizontal );
              maximizeChange();
              break;
          default:
              maximize(maximizeMode() == MaximizeFull ? MaximizeRestore : MaximizeFull );
              maximizeChange();
        }

        m_buttons[ButtonMaximize]->setMaximized( maximizeMode()!=MaximizeRestore );
        QToolTip::add(m_buttons[ButtonMaximize], (maximizeMode()!=MaximizeRestore) ? i18n("Restore") : i18n("Maximize"));
    }
}

//
// mousePosition(const QPoint &point) const
//
KDecoration::Position RidgeClient::mousePosition(const QPoint &point) const
{
	return KDecoration::mousePosition(point);
}

//
// minimumSize() const
//
QSize RidgeClient::minimumSize() const
{
    return QSize (100, 50);
}

//
// iconChange()
//
void RidgeClient::iconChange()
{
    if (hasMenu)
    {
	    widget()->repaint(topSpacer_->geometry(), false);
    }
}

//
// shadeChange ()
//
void RidgeClient::shadeChange ()
{
    widget()->repaint(topSpacer_->geometry(), false);
}

//
// resizeEvent(QResizeEvent *)
//
void RidgeClient::resizeEvent(QResizeEvent *)
{
    widget()->repaint();
}

//
// showEvent(QShowEvent *)
//
void RidgeClient::showEvent(QShowEvent *)
{
    widget()->repaint();
}

//
// isTool() const
//
bool RidgeClient::isTool() const
{
    const int SUPPORTED_WINDOW_TYPES_MASK =
        NET::NormalMask | NET::DesktopMask | NET::DockMask | NET::ToolbarMask |
        NET::MenuMask | NET::DialogMask | NET::OverrideMask | NET::UtilityMask |
        NET::SplashMask;
        NET::WindowType type = windowType (SUPPORTED_WINDOW_TYPES_MASK);
    return type == NET::Toolbar || type == NET::Utility || type == NET::Menu;
}

